/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include <stdio.h>
#include "ap_int.h"
#include "hls_interface.h"
#include "hls_utils.h"
#include "hls_aes.h"

ap_uint<8> xtime2(ap_uint<8> a)
{
#pragma HLS INLINE
  if ((a >> 7) == 1)
    return ((a << 1) ^ 0x1b);
  else
    return (a << 1);
}
ap_uint<8> xtime3(ap_uint<8> a)
{
#pragma HLS INLINE
  return (xtime2(a)^a);
}
ap_uint<8> xtime4(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime2(xtime2(a));
}
ap_uint<8> xtime8(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime4(xtime2(a));
}
ap_uint<8> xtime9(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^a;
}
ap_uint<8> xtimeb(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime2(a)^a;
}
ap_uint<8> xtimed(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime4(a)^a;
}
ap_uint<8> xtimee(ap_uint<8> a)
{
#pragma HLS INLINE
  return xtime8(a)^xtime4(a)^xtime2(a);
}

void ShiftRows(array_t ostate, array_t istate)
{
#pragma HLS INLINE
  ostate[0]  = istate[0];
  ostate[4]  = istate[4];
  ostate[8]  = istate[8];
  ostate[12] = istate[12];

  ostate[13] = istate[1];
  ostate[1]  = istate[5];
  ostate[5]  = istate[9];
  ostate[9]  = istate[13];

  ostate[10] = istate[2];
  ostate[2]  = istate[10];
  ostate[14] = istate[6];
  ostate[6]  = istate[14];

  ostate[7]  = istate[3];
  ostate[3]  = istate[15];
  ostate[15] = istate[11];
  ostate[11] = istate[7];
}

void iShiftRows(array_t ostate, array_t istate)
{
#pragma HLS INLINE
  ostate[0]  = istate[0];
  ostate[4]  = istate[4];
  ostate[8]  = istate[8];
  ostate[12] = istate[12];

  ostate[13] = istate[9];
  ostate[1]  = istate[13];
  ostate[5]  = istate[1];
  ostate[9]  = istate[5];

  ostate[10] = istate[2];
  ostate[2]  = istate[10];
  ostate[14] = istate[6];
  ostate[6]  = istate[14];

  ostate[3]  = istate[7];
  ostate[7]  = istate[11];
  ostate[11] = istate[15];
  ostate[15] = istate[3];
}

void SubBytes(array_t ostate, array_t istate)
{
#pragma HLS INLINE
#pragma HLS RESOURCE variable=Sbox00 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox01 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox02 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox03 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox04 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox05 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox06 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox07 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox08 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox09 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox10 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox11 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox12 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox13 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox14 core=ROM_1P_1S
#pragma HLS RESOURCE variable=Sbox15 core=ROM_1P_1S

  ostate[0]  = Sbox00[istate[0]];
  ostate[1]  = Sbox01[istate[1]];
  ostate[2]  = Sbox02[istate[2]];
  ostate[3]  = Sbox03[istate[3]];
  ostate[4]  = Sbox04[istate[4]];
  ostate[5]  = Sbox05[istate[5]];
  ostate[6]  = Sbox06[istate[6]];
  ostate[7]  = Sbox07[istate[7]];
  ostate[8]  = Sbox08[istate[8]];
  ostate[9]  = Sbox09[istate[9]];
  ostate[10] = Sbox10[istate[10]];
  ostate[11] = Sbox11[istate[11]];
  ostate[12] = Sbox12[istate[12]];
  ostate[13] = Sbox13[istate[13]];
  ostate[14] = Sbox14[istate[14]];
  ostate[15] = Sbox15[istate[15]];
}

void iSubBytes(array_t ostate, array_t istate)
{
#pragma HLS INLINE
#pragma HLS RESOURCE variable=iSbox00 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox01 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox02 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox03 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox04 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox05 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox06 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox07 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox08 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox09 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox10 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox11 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox12 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox13 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox14 core=ROM_1P_1S
#pragma HLS RESOURCE variable=iSbox15 core=ROM_1P_1S

  ostate[0]  = iSbox00[istate[0]];
  ostate[1]  = iSbox01[istate[1]];
  ostate[2]  = iSbox02[istate[2]];
  ostate[3]  = iSbox03[istate[3]];
  ostate[4]  = iSbox04[istate[4]];
  ostate[5]  = iSbox05[istate[5]];
  ostate[6]  = iSbox06[istate[6]];
  ostate[7]  = iSbox07[istate[7]];
  ostate[8]  = iSbox08[istate[8]];
  ostate[9]  = iSbox09[istate[9]];
  ostate[10] = iSbox10[istate[10]];
  ostate[11] = iSbox11[istate[11]];
  ostate[12] = iSbox12[istate[12]];
  ostate[13] = iSbox13[istate[13]];
  ostate[14] = iSbox14[istate[14]];
  ostate[15] = iSbox15[istate[15]];
}

void MixColumns(array_t ostate, array_t istate)
{
#pragma HLS INLINE
    ostate[0]  = xtime2(istate[ 0]) ^ xtime3(istate[ 1]) ^        istate[ 2]  ^        istate[ 3] ;
    ostate[1]  =        istate[ 0]  ^ xtime2(istate[ 1]) ^ xtime3(istate[ 2]) ^        istate[ 3] ;
    ostate[2]  =        istate[ 0]  ^        istate[ 1]  ^ xtime2(istate[ 2]) ^ xtime3(istate[ 3]);
    ostate[3]  = xtime3(istate[ 0]) ^        istate[ 1]  ^        istate[ 2]  ^ xtime2(istate[ 3]);

    ostate[4]  = xtime2(istate[ 4]) ^ xtime3(istate[ 5]) ^        istate[ 6]  ^        istate[ 7] ;
    ostate[5]  =        istate[ 4]  ^ xtime2(istate[ 5]) ^ xtime3(istate[ 6]) ^        istate[ 7] ;
    ostate[6]  =        istate[ 4]  ^        istate[ 5]  ^ xtime2(istate[ 6]) ^ xtime3(istate[ 7]);
    ostate[7]  = xtime3(istate[ 4]) ^        istate[ 5]  ^        istate[ 6]  ^ xtime2(istate[ 7]);

    ostate[8]  = xtime2(istate[ 8]) ^ xtime3(istate[ 9]) ^        istate[10]  ^        istate[11] ;
    ostate[9]  =        istate[ 8]  ^ xtime2(istate[ 9]) ^ xtime3(istate[10]) ^        istate[11] ;
    ostate[10] =        istate[ 8]  ^        istate[ 9]  ^ xtime2(istate[10]) ^ xtime3(istate[11]);
    ostate[11] = xtime3(istate[ 8]) ^        istate[ 9]  ^        istate[10]  ^ xtime2(istate[11]);

    ostate[12] = xtime2(istate[12]) ^ xtime3(istate[13]) ^        istate[14]  ^        istate[15] ;
    ostate[13] =        istate[12]  ^ xtime2(istate[13]) ^ xtime3(istate[14]) ^        istate[15] ;
    ostate[14] =        istate[12]  ^        istate[13]  ^ xtime2(istate[14]) ^ xtime3(istate[15]);
    ostate[15] = xtime3(istate[12]) ^        istate[13]  ^        istate[14]  ^ xtime2(istate[15]);
}

void iMixColumns(array_t ostate, array_t istate)
{
#pragma HLS INLINE
    ostate[0]  = xtimee(istate[ 0]) ^ xtimeb(istate[ 1]) ^ xtimed(istate[ 2]) ^ xtime9(istate[ 3]);
    ostate[1]  = xtime9(istate[ 0]) ^ xtimee(istate[ 1]) ^ xtimeb(istate[ 2]) ^ xtimed(istate[ 3]);
    ostate[2]  = xtimed(istate[ 0]) ^ xtime9(istate[ 1]) ^ xtimee(istate[ 2]) ^ xtimeb(istate[ 3]);
    ostate[3]  = xtimeb(istate[ 0]) ^ xtimed(istate[ 1]) ^ xtime9(istate[ 2]) ^ xtimee(istate[ 3]);

    ostate[4]  = xtimee(istate[ 4]) ^ xtimeb(istate[ 5]) ^ xtimed(istate[ 6]) ^ xtime9(istate[ 7]);
    ostate[5]  = xtime9(istate[ 4]) ^ xtimee(istate[ 5]) ^ xtimeb(istate[ 6]) ^ xtimed(istate[ 7]);
    ostate[6]  = xtimed(istate[ 4]) ^ xtime9(istate[ 5]) ^ xtimee(istate[ 6]) ^ xtimeb(istate[ 7]);
    ostate[7]  = xtimeb(istate[ 4]) ^ xtimed(istate[ 5]) ^ xtime9(istate[ 6]) ^ xtimee(istate[ 7]);

    ostate[8]  = xtimee(istate[ 8]) ^ xtimeb(istate[ 9]) ^ xtimed(istate[10]) ^ xtime9(istate[11]);
    ostate[9]  = xtime9(istate[ 8]) ^ xtimee(istate[ 9]) ^ xtimeb(istate[10]) ^ xtimed(istate[11]);
    ostate[10] = xtimed(istate[ 8]) ^ xtime9(istate[ 9]) ^ xtimee(istate[10]) ^ xtimeb(istate[11]);
    ostate[11] = xtimeb(istate[ 8]) ^ xtimed(istate[ 9]) ^ xtime9(istate[10]) ^ xtimee(istate[11]);

    ostate[12] = xtimee(istate[12]) ^ xtimeb(istate[13]) ^ xtimed(istate[14]) ^ xtime9(istate[15]);
    ostate[13] = xtime9(istate[12]) ^ xtimee(istate[13]) ^ xtimeb(istate[14]) ^ xtimed(istate[15]);
    ostate[14] = xtimed(istate[12]) ^ xtime9(istate[13]) ^ xtimee(istate[14]) ^ xtimeb(istate[15]);
    ostate[15] = xtimeb(istate[12]) ^ xtimed(istate[13]) ^ xtime9(istate[14]) ^ xtimee(istate[15]);
}

void AddRoundKey(array_t ostate, array_t istate, array_t rk)
{
#pragma HLS INLINE
    xor2(ostate, istate, rk);
}

void AESRound(array_t ostate, array_t ostate_fdb, array_t istate, array_t rk)
{
    array_t sub, shift, mix, add;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate_fdb         complete dim=1
#pragma HLS ARRAY_RESHAPE variable=sub                complete dim=1
#pragma HLS ARRAY_RESHAPE variable=shift              complete dim=1
#pragma HLS ARRAY_RESHAPE variable=mix                complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add                complete dim=1

    SubBytes(sub, istate);
    ShiftRows(shift, sub);
    MixColumns(mix, shift);
    AddRoundKey(add, mix, rk);

    copy(ostate_fdb, add);
    copy(ostate, shift);
}

void AESInvRound(array_t ostate, array_t ostate_fdb, array_t istate, array_t rk)
{
    array_t isub, ishift, imix, add;
#pragma HLS ARRAY_RESHAPE variable=istate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=rk                 complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ostate_fdb         complete dim=1
#pragma HLS ARRAY_RESHAPE variable=isub               complete dim=1
#pragma HLS ARRAY_RESHAPE variable=ishift             complete dim=1
#pragma HLS ARRAY_RESHAPE variable=imix               complete dim=1
#pragma HLS ARRAY_RESHAPE variable=add                complete dim=1

    iShiftRows(ishift, istate);
    iSubBytes(isub, ishift);
    AddRoundKey(add, isub, rk);
    iMixColumns(imix, add);
  
    copy(ostate_fdb, imix);
    copy(ostate, add);
}


void AESKeyUpdate(array_t new_rk, array_t rk, ap_uint<4> round)
{
#pragma HLS INLINE
    int i, j;
    ap_uint<8> tk[16];

    static const ap_uint<8> rcon[10] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36};

#pragma HLS RESOURCE variable=SboxK0 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK1 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK2 core=ROM_1P_1S
#pragma HLS RESOURCE variable=SboxK3 core=ROM_1P_1S
#pragma HLS RESOURCE variable=rcon   core=ROM_1P_1S
#pragma HLS ARRAY_RESHAPE variable=tk complete dim=1 reshape

    for(j = 0; j < 4; j++)
    {
#pragma HLS UNROLL
      for(i = 0; i < 4; i++)
      {
#pragma HLS UNROLL
        tk[i+j*4] = rk[i+j*4];
      }
    }

    /* while not enough round key material calculated */
    /* calculate new values */
    tk[0] ^= SboxK0[tk[1+4*3]] ^ rcon[round];
    tk[1] ^= SboxK1[tk[2+4*3]];
    tk[2] ^= SboxK2[tk[3+4*3]];
    tk[3] ^= SboxK3[tk[0+4*3]];

    for(j = 1; j < 4; j++)
#pragma HLS UNROLL
        for(i = 0; i < 4; i++)
#pragma HLS UNROLL
          tk[i+4*j] ^= tk[i+4*(j-1)];

    /* copy values into round key array */
    for(j = 0; j < 4; j++)
#pragma HLS UNROLL
      for(i = 0; i < 4; i++)
#pragma HLS UNROLL
        new_rk[i+4*j] = tk[i+4*j];
}
