/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#include "ap_int.h"
#include "hls_stream.h"

#ifndef HLS_INTERFACE_H
#define HLS_INTERFACE_H

#define BLOCKSIZE 128
#define WORDSIZE  8
#define ELEMENTS  BLOCKSIZE/WORDSIZE


#define TYPE_AD 0
#define TYPE_D  2
#define TYPE_LEN 4
#define TYPE_TAG 5
#define TYPE NPUB 6
#define TYPE_NSEC 7

typedef ap_uint<BLOCKSIZE>    data_t;
typedef ap_uint<BLOCKSIZE/8>  data_bytes_t;
typedef ap_uint<5>   bsize_t;
typedef ap_uint<WORDSIZE> array_t [ELEMENTS];

struct public_bus {    
    data_t          data;
    data_bytes_t    valid_bytes;
    data_bytes_t    pad_loc;
    bsize_t         size;
    ap_uint<3>      type;
    bool            eoi;
    bool            eot;
    bool            partial;
};

struct output_bus {
    data_t          data;
    bsize_t         size;
};

struct secret_bus {
    data_t          data;
};

typedef hls::stream<public_bus> BDI_STREAM;
typedef hls::stream<secret_bus> KEY_STREAM;
typedef hls::stream<output_bus> BDO_STREAM;




void hls_aes_gcm (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    bool key_update
    );


#define CALL_CRYPTO hls_aes_gcm( \
        bdo_stream, &msg_auth_valid, \
        bdi_stream, key_stream, \
        decrypt, key_update);

#endif /* HLS_INTERFACE_H */
